#ifndef _DAWARE_H_
#define _DAWARE_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <signal.h>
#include <syslog.h>


extern char*
daware_get_container_name(char *name, size_t len);


typedef long DTermination;

extern DTermination*
daware_termination_register_callback(int signo, int (*func)(int, void*), void *arg);
extern void
daware_termination_unregister_callback(DTermination *dterm);


extern void
daware_log_open(const char *ident, int option, int facility);

extern void
daware_log_close();

extern void
daware_log_set_format(const char *logfmt);

extern void
daware_log(int priority, const char *file,
             int line, const char *func, const char *fmt, ...);

#define DLOG(priority, fmt, ...)                       \
    daware_log(priority, __FILE__, __LINE__, __func__, \
               fmt, ##__VA_ARGS__)

#define DLOG_EMERG(fmt, ...) \
    DLOG(LOG_EMERG, fmt, ##__VA_ARGS__);

#define DLOG_ALERT(fmt, ...) \
    DLOG(LOG_ALERT, fmt, ##__VA_ARGS__);

#define DLOG_CRIT(fmt, ...) \
    DLOG(LOG_CRIT, fmt, ##__VA_ARGS__);

#define DLOG_ERR(fmt, ...) \
    DLOG(LOG_ERR, fmt, ##__VA_ARGS__);

#define DLOG_WARNING(fmt, ...) \
    DLOG(LOG_WARNING, fmt, ##__VA_ARGS__);

#define DLOG_NOTICE(fmt, ...) \
    DLOG(LOG_NOTICE, fmt, ##__VA_ARGS__);

#define DLOG_INFO(fmt, ...) \
    DLOG(LOG_INFO, fmt, ##__VA_ARGS__);

#define DLOG_DEBUG(fmt, ...) \
    DLOG(LOG_DEBUG, fmt, ##__VA_ARGS__);

#ifdef __cplusplus
}
#endif

#endif /*_DAWARE_H*/
