/*
 * (C) Copyright 2011-2013 JST - Japan Science and Technology Agency
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>

#define SYSLOG_NAMES //for facilitynames and prioritynames in syslog.h
#include <syslog.h>

#include "daware.h"

void
usage()
{
    fprintf(stderr,
            "usage: dre-logger [-is] [-f format] [-p pri] [-t tag] [ message ... ]\n");
    exit(1);
}

CODE*
find_code_by_name(CODE *codes, char *name)
{
    CODE *c;

    for (c = codes; c->c_name; c++)
        if (!strcasecmp(name, c->c_name))
            return c;

    return NULL;
}

int
get_pricode(char *str)
{
    char *p;
    int fac, pri;
    CODE *code;

    for (p = str; *p && *p != '.'; p++);
    if (*p == '.') {
        *p = '\0';
        if (isdigit(*str)) {
            fac = atoi(str);
        } else {
            if ((code = find_code_by_name(facilitynames, str)) == NULL) {
                fprintf(stderr, "unknown facility name: %s.\n", str);
                exit(1);
            }
            fac = code->c_val;
        }

        *p++ = '.';
        if (isdigit(*p)) {
            pri = atoi(p);
        } else {
            code = find_code_by_name(prioritynames, p);
            if ((code = find_code_by_name(prioritynames, p)) == NULL) {
                fprintf(stderr, "unknown priority name: %s.\n", p);
                exit(1);
            }
            pri = code->c_val;
        }
    } else {
        fac = LOG_LOCAL5;
        if (isdigit(*str)) {
            pri = atoi(str);
        } else {
            code = find_code_by_name(prioritynames, str);
            if ((code = find_code_by_name(prioritynames, str)) == NULL) {
                fprintf(stderr, "unknown priority name: %s.\n", str);
                exit(1);
            }
            pri = code->c_val;
        }
    }
    return ((pri & LOG_PRIMASK) | (fac & LOG_FACMASK));
}

int main(int argc, char * const argv[])
{
    int ch, logflags, pri;
    char *tag, *fmt, buf[10];

    tag = NULL;
    fmt = NULL;
    pri = LOG_NOTICE;
    logflags = 0;
    while ((ch = getopt(argc, argv, "ip:st:f:")) != -1) {
        switch((char)ch) {
        case 'i':
            logflags |= LOG_PID;
            break;
        case 'p':
            pri = get_pricode(optarg);
            break;
        case 's':
            logflags |= LOG_PERROR;
            break;
        case 't':
            tag = optarg;
            break;
        case 'f':
            fmt = optarg;
            break;
        case '?':
        default:
            usage();
        }
    }
    argc -= optind;
    argv += optind;

    if (fmt) daware_log_set_format(fmt);

    daware_log_open(tag ? tag : getlogin(), logflags, 0);

    if (argc > 0) {
        char *p;
        int max, len1, len2;

        max = sizeof(buf) - 1;
        buf[0] = '\0';

        while (*argv) {
            len1 = strlen(buf);
            len2 = strlen(*argv);

            if (max < len1 + len2 + 1 && 0 < len1) {
                daware_log(pri, NULL, 0, NULL, "%s", buf);
                buf[0] = '\0';
                len1 = 0;
            }

            if (max < len2) {
                daware_log(pri, NULL, 0, NULL, "%s", *argv);
            } else {
                if (len1 != 0)
                    strcat(buf, " ");
                strcat(buf, *argv);
            }

            *argv++;
        }
        if (0 < strlen(buf)) {
            daware_log(pri, NULL, 0, NULL, "%s", buf);
        }
    }

    daware_log_close();

    exit(0);
}
