/*
 * (C) Copyright 2009-2013 JST - Japan Science and Technology Agency
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include "dre_app.h"

#ifndef DRE_BUFFER_LENGTH
#define DRE_BUFFER_LENGTH 1024
#endif


/*
 * escape meta charactors(local function)
 */
static dre_return_t
escape_meta_char(char *dest, const char *src)
{
    dre_app_return_t ret = {0};
    char *p1, *p2;
    char tmp1[DRE_BUFFER_LENGTH] = "";
    char tmp2[DRE_BUFFER_LENGTH] = "";
    int c = 0;
    char meta[] = {'\\', '"', '$', '@', '\'', '(', ')', '^', '|', 
                   '[', ']', '{', '}', ';', '*', '?'};

    if(!src || !dest) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    strcpy(dest, "");
    strcpy(tmp1, src);
    for(c = 0; c < sizeof(meta); c++) {
        char escaped[3];
        p1 = tmp1;
        sprintf(escaped, "\\%c", meta[c]);

        while(1) {
            p2 = strchr(p1, meta[c]);
            if(p2) {
                if(strlen(tmp2) + strlen(p1) - strlen(p2) < DRE_BUFFER_LENGTH) {
                    strncat(tmp2, p1, strlen(p1) - strlen(p2));
                } else {
                    ret.error = DRE_ERR_REQUEST_TOO_LONG;
                    goto _DRE_EXIT_;
                }

                if(strlen(tmp2) + strlen(escaped) < DRE_BUFFER_LENGTH) {
                    strcat(tmp2, escaped);
                } else {
                    ret.error = DRE_ERR_REQUEST_TOO_LONG;
                    goto _DRE_EXIT_;
                }
                p1 += strlen(p1) - strlen(p2) + sizeof(char);
            } else {
                if(strlen(tmp2) + strlen(p1) < DRE_BUFFER_LENGTH) {
                    strcat(tmp2, p1);
                } else {
                    ret.error = DRE_ERR_REQUEST_TOO_LONG;
                    goto _DRE_EXIT_;
                }
                break;
            }
        }
        strcpy(tmp1, tmp2);
        strcpy(tmp2, "");
    }
    strcpy(dest, tmp1);

_DRE_EXIT_:
    return ret.value;
}


/*
 * create container
 */
dre_return_t
dre_app_create_container(const char *name, 
                         const dre_app_data *data)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app create -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, name);

    switch(data->type) {
    case DRE_APP_CONTAINER_TYPE_LXC:
        if(data->udata.lxc.port > 0) {
            snprintf(option, DRE_BUFFER_LENGTH, " -p %d", data->udata.lxc.port);
            if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
                ret.error = DRE_ERR_REQUEST_TOO_LONG;
                goto _DRE_EXIT_;
            }
            strcat(cmd, option);
        }
        if(data->udata.lxc.base && strlen(data->udata.lxc.base)) {
            if(strlen(data->udata.lxc.base) > DRE_BUFFER_LENGTH) {
                ret.error = DRE_ERR_REQUEST_TOO_LONG;
                goto _DRE_EXIT_;
            }
            snprintf(option, DRE_BUFFER_LENGTH, " -b %s", data->udata.lxc.base);
            if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
                ret.error = DRE_ERR_REQUEST_TOO_LONG;
                goto _DRE_EXIT_;
            }
            strcat(cmd, option);
        }
        break;
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: create: Failed to create '%s'. (ErrNo=%d)", name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: create: Created application container '%s'.", name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * destroy container
 */
dre_return_t
dre_app_destroy_container(const char *name)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app destroy -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: destroy: Failed to destroy '%s'. (ErrNo=%d)", name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: destroy: Destroyed application container '%s'.", name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * open handler
 */
dre_app_handle *
dre_app_open(const char *name)
{
    dre_err_t ret = 0;
    dre_app_handle *h = NULL;
    char cmd[DRE_BUFFER_LENGTH] = "";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char buf[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";
    FILE *p = NULL;

    /* search container */
    strcpy(cmd, "sudo dre-app search -n ");
    if(strlen(cmd) + strlen(name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, name);
    strcat(cmd, redirect);
    escape_meta_char(escaped_cmd, cmd);

    ret = -WEXITSTATUS(system(escaped_cmd));
    if(!ret) {
        h = (dre_app_handle *)malloc(sizeof(dre_app_handle));
        h->name = (char *)malloc(sizeof(char) * strlen(name));
        strcpy(h->name, name);
        h->type = 0;
    } else {
        ret = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    /* open handler */
    strcpy(cmd, "sudo dre-app info -n ");
    if(strlen(cmd) + strlen(name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    p = popen(escaped_cmd, "r");
    if(p) {
        while(fgets(buf, sizeof(buf), p)) {
            if(strstr(buf, "CONTAINER_TYPE")) {
                char *tmp;
                tmp = strtok(buf, " ");
                h->type = atoi(strtok(NULL, " "));
            }
            break;
        }
    } else {
        ret = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

_DRE_EXIT_:
    if(p) {
        pclose(p);
    }

    if(!h || !(h->type)) {
        DLOG_ERR("DRE-APP-LIB: open: Failed to open handler of '%s'. (ErrNo=%d)", name, ret);
    } else {
        DLOG_INFO("DRE-APP-LIB: open: Opened handler of '%s'.", name);
    }

    return h;
}


/*
 * start container
 */
dre_return_t
dre_app_start(const dre_app_handle *h)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app start -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: start: Failed to start '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: start: Started application container '%s'.", h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * stop container
 */
dre_return_t
dre_app_stop(const dre_app_handle *h)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app stop -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: stop: Failed to stop '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: stop: Stopped application container '%s'.", h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * close handler
 */
dre_return_t
dre_app_close(dre_app_handle *h)
{
    dre_app_return_t ret = {0};
    char name[DRE_BUFFER_LENGTH] = "";

    strcpy(name, h->name);

    if(h) {
        if(h->name) {
            free(h->name);
            h->name = NULL;
        } else {
            ret.error = DRE_ERR_NULL_POINTER;
            goto _DRE_EXIT_;
        }
        free(h);
        h = NULL;
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

_DRE_EXIT_:
    if(h) {
        free(h);
        h = NULL;
    }

    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: close: Failed to close handler of '%s'. (ErrNo=%d)", name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: close: Closed handler of '%s'.", name);
    }

    return ret.value;
}


/*
 * get status
 */
dre_return_t
dre_app_get_status(const dre_app_handle *h)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app status -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";
    int result = 0;

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    result = WEXITSTATUS(system(escaped_cmd));
    if(result >= -DRE_ERR_INVALID_ARGUMENT) {
        ret.error = -result;
    } else {
        ret.status = result;
    }

    if(ret.value < 0) {
        DLOG_ERR("DRE-APP-LIB: status: Failed to get status of '%s'. (ErrNo=%d)", h->name, ret.value);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * get pid of container
 */
dre_return_t
dre_app_get_pid(const dre_app_handle *h, int *pid)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app getpid -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char buf[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";
    FILE *p;

    if(!h || !pid) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    p = popen(escaped_cmd, "r");
    if(p) {
        char *ps = fgets(buf, sizeof(buf), p);
        if(!ps) {
            // TODO: error
        }
        ret.error = -WEXITSTATUS(pclose(p));
        if(!ret.error) {
            *pid = atoi(buf);
        }
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

_DRE_EXIT_:
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: getpid: Failed to get PID of '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: getpid: PID of '%s' is %d.", h->name, *pid);
    }

    return ret.value;
}


/*
 * create control group
 */
dre_return_t
dre_app_create_cgroup(const dre_app_handle *h, const char *cgroup)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app create -c ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(cgroup) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, cgroup);

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: create: Failed to create control group '%s'. (ErrNo=%d)", cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: create: Created control group '%s'.", cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * destroy control group
 */
dre_return_t
dre_app_destroy_cgroup(const dre_app_handle *h, const char *cgroup)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app destroy -g ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(cgroup) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, cgroup);

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: destroy: Failed to destroy control group '%s'. (ErrNo=%d)", cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: destroy: Destroyed control group '%s'.", cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}

#if 0
/*
 * bind container to control group
 */
dre_return_t
dre_app_bind_container(const dre_app_handle *h, const char *cgroup)
{
    dre_app_return_t ret = {0};
    int pid;

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    ret.value = dre_app_get_pid(h, &pid);
    if(ret.value) {
        goto _DRE_EXIT_;
    }

    ret.value = dre_app_bind_pid(NULL, cgroup, pid);
    if(ret.value) {
        goto _DRE_EXIT_;
    }

    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: bind: Failed to bind '%s' to '%s'. (ErrNo=%d)", h->name, cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: bind: Bound '%s' to '%s'.", h->name, cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}
#endif

/*
 * bind pid to control group
 */
dre_return_t
dre_app_bind_pid(const dre_app_handle *h, const char *cgroup, const int pid)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app bind -g ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(cgroup) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, cgroup);

    sprintf(option, " -pid %d", pid);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: bind: Failed to bind '%d' to '%s'. (ErrNo=%d)", pid, cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: bind: Bound '%d' to '%s'.", pid, cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * bind vpid to control group
 */
dre_return_t
dre_app_bind_vpid(const dre_app_handle *h, const char *cgroup, const int vpid)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app bind -g ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(cgroup) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, cgroup);

    sprintf(option, " -vpid %d", vpid);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: bind: Failed to bind '%d' to '%s'. (ErrNo=%d)", vpid, cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: bind: Bound '%d' to '%s'.", vpid, cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}

#if 0
/*
 * unbind container from control group
 */
dre_return_t
dre_app_unbind_container(const dre_app_handle *h, const char *cgroup)
{
    dre_app_return_t ret = {0};
    int pid;

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    ret.value = dre_app_get_pid(h, &pid);
    if(ret.value) {
        goto _DRE_EXIT_;
    }

    ret.value = dre_app_unbind_pid(NULL, cgroup, pid);
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: unbind: Failed to unbind '%s' from '%s'. (ErrNo=%d)", h->name, cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: unbind: Unbound '%s' from '%s'.", h->name, cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}
#endif

/*
 * unbind pid from control group
 */
dre_return_t
dre_app_unbind_pid(const dre_app_handle *h, const char *cgroup, const int pid)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app unbind -c ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(cgroup) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, cgroup);

    sprintf(option, " -pid %d", pid);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: unbind: Failed to unbind '%d' from '%s'. (ErrNo=%d)", pid, cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: unbind: Unbound '%d' from '%s'.", pid, cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * unbind vpid from control group
 */
dre_return_t
dre_app_unbind_vpid(const dre_app_handle *h, const char *cgroup, const int vpid)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app unbind -g ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(!cgroup) {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(strlen(cmd) + strlen(cgroup) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, cgroup);

    sprintf(option, " -vpid %d", vpid);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(strlen(cmd) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: unbind: Failed to unbind '%d' from '%s'. (ErrNo=%d)", vpid, cgroup, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: unbind: Unbound '%d' from '%s'.", vpid, cgroup);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * register pid
 */
dre_return_t
dre_app_register_pid(const dre_app_handle *h, const int pid)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app register -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -pid %d", pid);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: register: Failed to register '%d'. (ErrNo=%d)", pid, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: register: Registered '%d'.", pid);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * register vpid
 */
dre_return_t
dre_app_register_vpid(const dre_app_handle *h, const int vpid)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app register -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -vpid %d", vpid);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: register: Failed to register '%d'. (ErrNo=%d)", vpid, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: register: Registered '%d'.", vpid);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * unregister pid
 */
dre_return_t
dre_app_unregister_pid(const dre_app_handle *h, const int pid)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app unregister -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -pid %d", pid);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: unregister: Failed to unregister '%d'. (ErrNo=%d)", pid, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: unregister: Unregistered '%d'.", pid);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * unregister vpid
 */
dre_return_t
dre_app_unregister_vpid(const dre_app_handle *h, const int vpid)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app unregister -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -vpid %d", vpid);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.status = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: unregister: Failed to unregister '%d'. (ErrNo=%d)", vpid, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: unregister: Unregistered '%d'.", vpid);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * get registered pid list
 */
dre_return_t
dre_app_get_registered_pid_list(const dre_app_handle *h, 
                                    dre_app_pid_list_t *list)
{
    dre_app_return_t ret = {0};
    int i = 0;
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app pidlist -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char buf[DRE_BUFFER_LENGTH] = "";
    dre_app_pid_list_t plist = NULL;
    dre_app_pid_t pid;
    FILE *p;

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    escape_meta_char(escaped_cmd, cmd);
    p = popen(escaped_cmd, "r");
    if(p) {
        while(fgets(buf, sizeof(buf), p)) {
            buf[strlen(buf) - 1] = '\0';
            int rc = sscanf(buf, "%d,%d", &pid.pid, &pid.vpid);

            if(rc == 2) {
                plist = (dre_app_pid_list_t)realloc(plist, sizeof(dre_app_pid_t) * (i + 1));
                *(plist + i) = (dre_app_pid_t *)malloc(sizeof(dre_app_pid_t));
                memcpy(*(plist + i), &pid, sizeof(dre_app_pid_t));
                i++;
            }
        }
        if(plist) {
            *(plist + i) = NULL;
        }
        *list = plist;
        pclose(p);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * release registered list
 */
dre_return_t
dre_app_release_registered_pid_list(dre_app_pid_list_t list)
{
    dre_app_return_t ret = {0};
    int i = 0;

    if(list) {
        while(*(list + i)) {
            free(*(list + i));
            *(list + i) = NULL;
            i++;
        }
        free(list);
        list = NULL;
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
    }

    return ret.value;
}


/*
 * get container list
 */
dre_return_t
dre_app_get_container_list(dre_list_t *list)
{
    dre_app_return_t ret = {0};
    int i = 0;
    char buf[DRE_BUFFER_LENGTH] = "";
    dre_list_t plist = NULL;
    FILE *p;

    p = popen("sudo dre-app list", "r");
    if(p) {
        while(fgets(buf, sizeof(buf), p)) {
            buf[strlen(buf) - 1] = '\0';
            plist = (dre_list_t)realloc(plist, sizeof(buf) * (i + 1));
            *(plist + i) = (char *)malloc(sizeof(buf));
            strcpy(*(plist + i), buf);
            i++;
        }
        if(plist) {
            free(*(plist + i));
            *(plist + i) = NULL;
        }
        *list = plist;
        pclose(p);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
    }

    return ret.value;
}


/*
 * release list
 */
dre_return_t
dre_app_release_container_list(dre_list_t list)
{
    dre_app_return_t ret = {0};
    int i = 0;

    if(list) {
        while(*(list + i)) {
            free(*(list + i));
            *(list + i) = NULL;
            i++;
        }
        free(list);
        list = NULL;
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
    }

    return ret.value;
}


/*
 * get base image list
 */
dre_return_t
dre_app_get_base_list(dre_list_t *list)
{
    dre_app_return_t ret = {0};
    int i = 0;
    char buf[DRE_BUFFER_LENGTH] = "";
    dre_list_t plist = NULL;
    FILE *p;

    p = popen("sudo dre-app list --base", "r");
    if(p) {
        while(fgets(buf, sizeof(buf), p)) {
            buf[strlen(buf) - 1] = '\0';
            plist = (dre_list_t)realloc(plist, sizeof(buf) * (i + 1));
            *(plist + i) = (char *)malloc(sizeof(buf));
            strcpy(*(plist + i), buf);
            i++;
        }
        if(plist) {
            free(*(plist + i));
            *(plist + i) = NULL;
        }
        *list = plist;
        pclose(p);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
    }

    return ret.value;
}


/*
 * release base image list
 */
dre_return_t
dre_app_release_base_list(dre_list_t list)
{
    return dre_app_release_container_list(list);
}


/*
 * remove base image
 */
dre_return_t
dre_app_remove_base(const char *base)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app remove --base ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    sprintf(option, "%s", base);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: remove: Failed to remove base image '%s'. (ErrNo=%d)", base, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: remove: Removed base image '%s'.", base);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * get resource usage
 */
dre_return_t
dre_app_get_resource(const dre_app_handle *h, const int pid, const char *cgroup,  
                               dre_app_resource_value_t *resource)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char buf[DRE_BUFFER_LENGTH] = "";
    FILE *p;

    switch(resource->type) {
    case DRE_APP_CONTAINER_RESOURCE_CPU:
        sprintf(cmd, "sudo dre-app usage --cpu");
        break;
    case DRE_APP_CONTAINER_RESOURCE_CPU_CORE:
        sprintf(cmd, "sudo dre-app usage --core");
        break;
    case DRE_APP_CONTAINER_RESOURCE_MEMORY:
        sprintf(cmd, "sudo dre-app usage --mem");
        break;
    case DRE_APP_CONTAINER_RESOURCE_MEMSW:
        sprintf(cmd, "sudo dre-app usage --memsw");
        break;
    }

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(cgroup) {
        sprintf(option, " -g %s", cgroup);
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    if(pid > 0) {
        if(h) {
            sprintf(option, " -vpid %d", pid);
        } else {
            sprintf(option, " -pid %d", pid);
        }
        if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(cmd, option);
    }

    escape_meta_char(escaped_cmd, cmd);
    p = popen(escaped_cmd, "r");
    if(p) {
        char *pr = fgets(buf, sizeof(buf), p);
        if(!pr) {
            // TODO: error
        }
        buf[strlen(buf) - 1] = '\0';
        ret.error = -WEXITSTATUS(pclose(p));
        if(!ret.error) {
            switch(resource->type) {
            case DRE_APP_CONTAINER_RESOURCE_CPU:
                resource->udata.cpu.percent = atoi(buf);
                break;
            case DRE_APP_CONTAINER_RESOURCE_CPU_CORE:
                resource->udata.cpu.cores = buf;
                break;
            case DRE_APP_CONTAINER_RESOURCE_MEMORY:
                resource->udata.mem.usage_in_bytes = atoi(buf);
                break;
            case DRE_APP_CONTAINER_RESOURCE_MEMSW:
                resource->udata.mem.memsw_usage_in_bytes = atoi(buf);
                break;
            }
        }
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

_DRE_EXIT_:
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: usage: Failed to get resource usage of '%s'. (ErrNo=%d)", h->name, ret.value);
    }

    return ret.value;
}


/*
 * set resource usage
 */
dre_return_t
dre_app_set_resource(const dre_app_handle *h, const char *cgroup, 
                               const dre_app_resource_value_t resource)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char target[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(target) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(target, option);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(cgroup) {
        sprintf(option, " -c %s", cgroup);
        if(strlen(target) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(target, option);
    }

    switch(resource.type) {
    case DRE_APP_CONTAINER_RESOURCE_CPU:
        sprintf(cmd, "sudo dre-app limit --cpu %d", resource.udata.cpu.percent);
        break;
    case DRE_APP_CONTAINER_RESOURCE_CPU_CORE:
        sprintf(cmd, "sudo dre-app limit --core %s", resource.udata.cpu.cores);
        break;
    case DRE_APP_CONTAINER_RESOURCE_MEMORY:
        sprintf(cmd, "sudo dre-app limit --mem %d", resource.udata.mem.usage_in_bytes);
        break;
    case DRE_APP_CONTAINER_RESOURCE_MEMSW:
        sprintf(cmd, "sudo dre-app limit --memsw %d", resource.udata.mem.memsw_usage_in_bytes);
        break;
    }

    if(strlen(cmd) + strlen(target) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }

    strcat(cmd, target);
    strcat(cmd, redirect);
    escape_meta_char(escaped_cmd, cmd);

    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value != 0) {
        goto _DRE_EXIT_;
    }

    DLOG_INFO("DRE-APP-LIB: limit: Set limitation of resource usage '%s'.", h->name);

_DRE_EXIT_:
    DLOG_ERR("DRE-APP-LIB: limit: Failed to set limitation of resource usage '%s'. (ErrNo=%d)", cgroup, ret.value);

    return ret.value;
}


/*
 * unset resource usage
 */
dre_return_t
dre_app_unset_resource(const dre_app_handle *h, const char *cgroup, 
                               const dre_app_resource_value_t resource)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char target[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(h) {
        sprintf(option, " -n %s", h->name);
        if(strlen(target) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(target, option);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

    if(cgroup) {
        sprintf(option, " -c %s", cgroup);
        if(strlen(target) + strlen(option) > DRE_BUFFER_LENGTH) {
            ret.error = DRE_ERR_REQUEST_TOO_LONG;
            goto _DRE_EXIT_;
        }
        strcat(target, option);
    }

    switch(resource.type) {
    case DRE_APP_CONTAINER_RESOURCE_CPU:
        sprintf(cmd, "sudo dre-app unlimit --cpu");
        break;
    case DRE_APP_CONTAINER_RESOURCE_CPU_CORE:
        sprintf(cmd, "sudo dre-app unlimit --core");
        break;
    case DRE_APP_CONTAINER_RESOURCE_MEMORY:
        sprintf(cmd, "sudo dre-app unlimit --mem");
        break;
    case DRE_APP_CONTAINER_RESOURCE_MEMSW:
        sprintf(cmd, "sudo dre-app unlimit --memsw");
        break;
    }

    if(strlen(cmd) + strlen(target) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }

    strcat(cmd, target);
    strcat(cmd, redirect);
    escape_meta_char(escaped_cmd, cmd);

    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value != 0) {
        goto _DRE_EXIT_;
    }

    DLOG_INFO("DRE-APP-LIB: unlimit: Unset limitation of resource usage '%s'.", h->name);

_DRE_EXIT_:
    DLOG_ERR("DRE-APP-LIB: unlimit: Failed to unset limitation of resource usage '%s'. (ErrNo=%d)", cgroup, ret.value);

    return ret.value;
}


/*
 * suspend container
 */
dre_return_t
dre_app_suspend(const dre_app_handle *h)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app suspend -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: suspend: Failed to suspend '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: suspend: Suspended '%s'.", h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * resume container
 */
dre_return_t
dre_app_resume(const dre_app_handle *h)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app resume -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: resume: Failed to resume '%s'. (ErrNo=%d)", h->name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: resume: Resumed '%s'.", h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * get checkpoint list
 */
dre_return_t
dre_app_get_checkpoint_list(const dre_app_handle *h, 
                                    dre_list_t *list)
{
    return 0;
}


/*
 * release checkpoint list
 */
dre_return_t
dre_app_release_checkpoint_list(dre_list_t list)
{
    return 0;
}


/*
 * create checkpoint
 */
dre_return_t
dre_app_checkpoint(const dre_app_handle *h, 
                             const char *tag)
{
    return 0;
}


/*
 * restart from checkpoint
 */
dre_return_t
dre_app_restart(const dre_app_handle *h, 
                          const char *tag)
{
    return 0;
}


/*
 * remove checkpoint
 */
dre_return_t
dre_app_remove_checkpoint(const dre_app_handle *h, 
                                  const char *tag)
{
    return 0;
}


/*
 * get snapshot list
 */
dre_return_t
dre_app_get_snapshot_list(const dre_app_handle *h, 
                                    dre_list_t *list)
{
    dre_app_return_t ret = {0};
    int i = 0;
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app list --snapshot -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char buf[DRE_BUFFER_LENGTH] = "";
    dre_list_t plist = NULL;
    FILE *p;

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    escape_meta_char(escaped_cmd, cmd);
    p = popen(escaped_cmd, "r");
    if(p) {
        if(h->type == DRE_APP_CONTAINER_TYPE_LXC) {
            while(fgets(buf, sizeof(buf), p)) {
                if(!strstr(buf, "TAG")) {
                    continue;
                } else {
                    break;
                }
            }
        }
        while(fgets(buf, sizeof(buf), p)) {
            char *tag;
            buf[strlen(buf) - 1] = '\0';

            if(h->type == DRE_APP_CONTAINER_TYPE_LXC) {
                tag = strtok(buf, " \t");
            } else {
                tag = buf;
            }

            plist = (dre_list_t)realloc(plist, sizeof(buf) * (i + 1));
            *(plist + i) = (char *)malloc(sizeof(buf));
            strcpy(*(plist + i), buf);
            i++;
        }

        if(plist) {
            *(plist + i) = NULL;
        }

        *list = plist;
        pclose(p);
    } else {
        ret.error = DRE_ERR_NULL_POINTER;
        goto _DRE_EXIT_;
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * release snapshot list
 */
dre_return_t
dre_app_release_snapshot_list(dre_list_t list)
{
    return dre_app_release_container_list(list);
}


/*
 * save snapshot
 */
dre_return_t
dre_app_save_snapshot(const dre_app_handle *h, 
                                const char *tag)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app save -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: save: Failed to save snapshot of '%s' as '%s'. (ErrNo=%d)", h->name, tag, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: save: Saved snapshot of '%s' as '%s'.", h->name, tag); 
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * load snapshot
 */
dre_return_t
dre_app_load_snapshot(const dre_app_handle *h, 
                                const char *tag)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app load -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: load: Failed to load snapshot '%s' of '%s'. (ErrNo=%d)", tag, h->name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: load: Loaded snapshot '%s' of '%s'.", tag, h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * remove snapshot
 */
dre_return_t
dre_app_remove_snapshot(const dre_app_handle *h, 
                                  const char *tag)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app remove --snapshot -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: remove: Failed to remove snapshot '%s' of '%s'. (ErrNo=%d)", tag, h->name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: remove: Removed snapshot '%s' of '%s'.", tag, h->name);
    }

_DRE_EXIT_:
    return ret.value;
}


/*
 * commit snapshot
 */
dre_return_t
dre_app_commit_snapshot(const dre_app_handle *h, 
                                  const char *tag, 
                                  const char *base)
{
    dre_app_return_t ret = {0};
    char cmd[DRE_BUFFER_LENGTH] = "sudo dre-app commit -n ";
    char escaped_cmd[DRE_BUFFER_LENGTH] = "";
    char option[DRE_BUFFER_LENGTH] = "";
    char *redirect = " > /dev/null 2>&1";

    if(strlen(cmd) + strlen(h->name) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, h->name);

    sprintf(option, " -t %s", tag);
    if(strlen(cmd) + strlen(option) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);

    sprintf(option, " -b %s", base);
    if(strlen(cmd) + strlen(option) + strlen(redirect) > DRE_BUFFER_LENGTH) {
        ret.error = DRE_ERR_REQUEST_TOO_LONG;
        goto _DRE_EXIT_;
    }
    strcat(cmd, option);
    strcat(cmd, redirect);

    escape_meta_char(escaped_cmd, cmd);
    ret.error = -WEXITSTATUS(system(escaped_cmd));
    if(ret.value) {
        DLOG_ERR("DRE-APP-LIB: commit: Failed to commit snapshot '%s' to '%s'. (ErrNo=%d)", tag, h->name, ret.value);
    } else {
        DLOG_INFO("DRE-APP-LIB: commit: Committed snapshot '%s' to '%s'.", tag, h->name);
    }

_DRE_EXIT_:
    return ret.value;
}

