/*
 * (C) Copyright 2009-2013 JST - Japan Science and Technology Agency
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifndef _DRE_SYS_H_
#define _DRE_SYS_H_

#include "dre_error.h"
#include "daware.h"

// list
#ifndef DRE_LIST
#define DRE_LIST
typedef char * dre_name_t;
typedef dre_name_t * dre_list_t;
#endif

#ifndef DRE_RETURN
#define DRE_RETURN
typedef int dre_return_t;
#endif


// container type
typedef enum {
    DRE_SYS_CONTAINER_TYPE_KVM = 100
} dre_sys_type_t;

// status type
typedef enum {
    DRE_SYS_CONTAINER_STATUS_STOP = 11,
    DRE_SYS_CONTAINER_STATUS_RUNNING, 
    DRE_SYS_CONTAINER_STATUS_SUSPENDED,
    DRE_SYS_CONTAINER_STATUS_BUSY
} dre_sys_status_t;

// return type
typedef union {
    dre_return_t value;
    dre_sys_status_t status;
    dre_err_t error;
} dre_sys_return_t;

// resource type
typedef enum {
    DRE_SYS_CONTAINER_RESOURCE_CPU = 1,
    DRE_SYS_CONTAINER_RESOURCE_CPU_CORE,
    DRE_SYS_CONTAINER_RESOURCE_MEMORY, 
    DRE_SYS_CONTAINER_RESOURCE_MEMSW 
} dre_sys_resource_t;


// container handler
typedef struct {
    dre_sys_type_t type;
    char *name;
} dre_sys_handle;

typedef dre_sys_handle dre_sys_target_handle;

// container data
typedef struct {
    dre_sys_type_t type;
    union {
        struct {
            int memory_size;
            const char *base;
            const char *rootfs;
        } kvm;
    } udata;
} dre_sys_data;

// resource usage
typedef struct {
    dre_sys_resource_t type;
    union {
        union {
            int percent;
            char *cores;
        } cpu;
        struct {
            int usage_in_bytes;
            int memsw_usage_in_bytes;
        } mem;
    } udata;
} dre_sys_resource_value_t;


#ifdef __cplusplus
extern "C" {
#endif

extern dre_return_t dre_sys_create_container(const char *name, const dre_sys_data *data);
extern dre_return_t dre_sys_destroy_container(const char *name);

extern dre_sys_handle *dre_sys_open(const char *name);
extern dre_return_t dre_sys_start(const dre_sys_handle *h);
extern dre_return_t dre_sys_stop(const dre_sys_handle *h);
extern dre_return_t dre_sys_close(dre_sys_handle *h);
extern dre_return_t dre_sys_get_status(const dre_sys_handle *h);
extern dre_return_t dre_sys_get_pid(const dre_sys_handle *h, int *pid);

extern dre_return_t dre_sys_create_cgroup(const dre_sys_handle *h, const char *cgroup);
extern dre_return_t dre_sys_destroy_cgroup(const dre_sys_handle *h, const char *cgroup);

extern dre_return_t dre_sys_bind_pid(const dre_sys_handle *h, const char *cgroup, const int pid);
extern dre_return_t dre_sys_unbind_pid(const dre_sys_handle *h, const char *cgroup, const int pid);

extern dre_return_t dre_sys_get_container_list(dre_list_t *list);
extern dre_return_t dre_sys_release_container_list(dre_list_t list);

extern dre_return_t dre_sys_get_base_list(dre_list_t *list);
extern dre_return_t dre_sys_release_base_list(dre_list_t list);
extern dre_return_t dre_sys_remove_base(const char *base);

extern dre_return_t dre_sys_immigration_start(const char *name, const int port);
extern dre_return_t dre_sys_migration_start(const dre_sys_handle *h, const char *target, const int port);

extern int dre_sys_get_resource(const dre_sys_handle *h, const int pid, const char *app, dre_sys_resource_value_t *resource);
extern dre_return_t dre_sys_set_resource(const dre_sys_handle *h, const char *cgroup, const char *app, const dre_sys_resource_value_t resource);
extern dre_return_t dre_sys_unset_resource(const dre_sys_handle *h, const char *cgroup, const char *app, const dre_sys_resource_value_t resource);

extern dre_return_t dre_sys_suspend(const dre_sys_handle *h);
extern dre_return_t dre_sys_resume(const dre_sys_handle *h);

extern dre_return_t dre_sys_get_checkpoint_list(const dre_sys_handle *h, dre_list_t *list);
extern dre_return_t dre_sys_release_checkpoint_list(dre_list_t list);
extern dre_return_t dre_sys_checkpoint(const dre_sys_handle *h, const char *tag);
extern dre_return_t dre_sys_restart(const dre_sys_handle *h, const char *tag);
extern dre_return_t dre_sys_remove_checkpoint(const dre_sys_handle *h, const char *tag);

extern dre_return_t dre_sys_share(const dre_sys_handle *h);
extern dre_return_t dre_sys_unshare(const dre_sys_handle *h);

extern dre_return_t dre_sys_get_snapshot_list(const dre_sys_handle *h, dre_list_t *list);
extern dre_return_t dre_sys_release_snapshot_list(dre_list_t list);
extern dre_return_t dre_sys_save_snapshot(const dre_sys_handle *h, const char *tag);
extern dre_return_t dre_sys_load_snapshot(const dre_sys_handle *h, const char *tag);
extern dre_return_t dre_sys_remove_snapshot(const dre_sys_handle *h, const char *tag);
extern dre_return_t dre_sys_commit_snapshot(const dre_sys_handle *h, const char *tag, const char *base);

#ifdef __cplusplus
}
#endif

#endif // _DRE_SYS_H_

