#include <stdio.h>
#include <stdlib.h>
#include "dre_sys.h"

int main(int argc, char **argv)
{
    dre_return_t result;
    dre_sys_handle *h = NULL;
    dre_sys_resource_value_t resource;
    char *name = NULL;
    char *cgroup = NULL;
    char *type = NULL;
    char *app = NULL;
    char *exec = argv[0];
    char **p = ++argv;

    int i;
    for(i = 0; p[i]; i++) {
        if(!strcmp(argv[i], "-n") && p[i + 1]) {
            name = argv[++i];
        } else if(!strcmp(argv[i], "--cpu") && p[i + 1]) {
           type = "cpu";
        } else if(!strcmp(argv[i], "--core") && p[i + 1]) {
           type = "core";
        } else if(!strcmp(argv[i], "--mem") && p[i + 1]) {
           type = "mem";
        } else if(!strcmp(argv[i], "--memsw") && p[i + 1]) {
           type = "memsw";
        } else if(!strcmp(argv[i], "-g") && p[i + 1]) {
            cgroup = argv[++i];
        } else if(!strcmp(argv[i], "-a") && p[i + 1]) {
            app = argv[++i];
        } else goto arg;
    }
    if(!name || !type) goto arg;

    /* unset resource usage */
    if(!strcmp(type, "cpu")) {
        resource.type = DRE_SYS_CONTAINER_RESOURCE_CPU;
    } else if(!strcmp(type, "core")) {
        resource.type = DRE_SYS_CONTAINER_RESOURCE_CPU_CORE;
    } else if(!strcmp(type, "mem")) {
        resource.type = DRE_SYS_CONTAINER_RESOURCE_MEMORY;
    } else if(!strcmp(type, "memsw")) {
        resource.type = DRE_SYS_CONTAINER_RESOURCE_MEMSW;
    } else {
        printf("Invalid resource type: %s\n", type);
        result = DRE_ERR_INVALID_ARGUMENT;
        goto error;
    }

    /* open handler */
    if(name) {
        h = dre_sys_open(name);
        if(!h) {
            result = DRE_ERR_NULL_POINTER;
            goto error;
        }
    }

    /* set limitation */
    result = dre_sys_unset_resource(h, cgroup, app, resource);
    if(result) goto error;
    printf("Unset limitation of %s.\n", type);

    /* close handler */
    if(h) {
        result = dre_sys_close(h);
        if(result) goto error;
    }

    return 0;

arg:
    printf("usage: %s -n <name> <resource_type> [-g <cgroup>] [-a <application_container>]\n", exec);
    return DRE_ERR_INVALID_ARGUMENT;

error:
    if(h) {
        dre_sys_close(h);
    }

    printf("Failed to set resource usage (ErrNo=%d).\n", result);
    return result;
}

